@@ -1,4 +1,4 @@
-/*  $Id: dbapi_impl_context.cpp 198023 2010-07-26 14:37:34Z ivanovp $
+/*  $Id: dbapi_impl_context.cpp 399230 2013-05-13 13:59:10Z ucko $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
@@ -65,6 +65,7 @@ namespace impl
 CDriverContext::CDriverContext(void) :
     m_LoginTimeout(0),
     m_Timeout(0),
+    m_CancelTimeout(0),
     m_MaxTextImageSize(0),
     m_ClientEncoding(eEncoding_ISO8859_1)
 {
@@ -132,7 +133,6 @@ unsigned int CDriverContext::GetTimeout(void) const
     return m_Timeout; 
 }
 
-
 bool CDriverContext::SetTimeout(unsigned int nof_secs)
 {
     bool success = true;
@@ -151,6 +151,27 @@ bool CDriverContext::SetTimeout(unsigned int nof_secs)
     return success;
 }
 
+unsigned int CDriverContext::GetCancelTimeout(void) const 
+{ 
+    CMutexGuard mg(m_CtxMtx);
+
+    return m_CancelTimeout;
+}
+
+bool CDriverContext::SetCancelTimeout(unsigned int nof_secs)
+{
+    bool success = true;
+    CMutexGuard mg(m_CtxMtx);
+
+    try {
+        m_CancelTimeout = nof_secs;
+    } catch (...) {
+        success = false;
+    }
+
+    return success;
+}
+
 bool CDriverContext::SetMaxTextImageSize(size_t nof_bytes)
 {
     CMutexGuard mg(m_CtxMtx);
@@ -271,40 +292,59 @@ void CDriverContext::CloseUnusedConnections(const string&   srv_name,
         if((!srv_name.empty()) && srv_name.compare(con->ServerName())) continue;
         if((!pool_name.empty()) && pool_name.compare(con->PoolName())) continue;
 
-        it = --m_NotInUse.erase(it);
+        it = m_NotInUse.erase(it);
+        --it;
         delete con;
     }
 }
 
-unsigned int CDriverContext::NofConnections(const string& srv_name,
-                                          const string& pool_name) const
+unsigned int CDriverContext::NofConnections(const TSvrRef& svr_ref,
+                                            const string& pool_name) const
 {
     CMutexGuard mg(m_CtxMtx);
 
-    if ( srv_name.empty() && pool_name.empty()) {
+    if ((!svr_ref  ||  !svr_ref->IsValid())  &&  pool_name.empty()) {
         return static_cast<unsigned int>(m_InUse.size() + m_NotInUse.size());
     }
 
-    int n = 0;
-    TConnPool::value_type con;
-
-    ITERATE(TConnPool, it, m_NotInUse) {
-        con = *it;
-        if((!srv_name.empty()) && srv_name.compare(con->ServerName())) continue;
-        if((!pool_name.empty()) && pool_name.compare(con->PoolName())) continue;
-        ++n;
+    string server;
+    Uint4 host = 0;
+    Uint2 port = 0;
+    if (svr_ref) {
+        host = svr_ref->GetHost();
+        port = svr_ref->GetPort();
+        if (host == 0)
+            server = svr_ref->GetName();
     }
 
-    ITERATE(TConnPool, it, m_InUse) {
-        con = *it;
-        if((!srv_name.empty()) && srv_name.compare(con->ServerName())) continue;
-        if((!pool_name.empty()) && pool_name.compare(con->PoolName())) continue;
-        ++n;
+    const TConnPool* pools[] = {&m_NotInUse, &m_InUse};
+    int n = 0;
+    for (size_t i = 0; i < ArraySize(pools); ++i) {
+        ITERATE(TConnPool, it, (*pools[i])) {
+            TConnPool::value_type con = *it;
+            if(!server.empty()) {
+                if (server.compare(con->ServerName()))
+                    continue;
+            }
+            else if (host != 0) {
+                if (host != con->Host()  ||  port != con->Port())
+                    continue;
+            }
+            if((!pool_name.empty()) && pool_name.compare(con->PoolName())) continue;
+            ++n;
+        }
     }
 
     return n;
 }
 
+unsigned int CDriverContext::NofConnections(const string& srv_name,
+                                            const string& pool_name) const
+{
+    TSvrRef svr_ref(new CDBServer(srv_name, 0, 0));
+    return NofConnections(svr_ref, pool_name);
+}
+
 CDB_Connection* CDriverContext::MakeCDBConnection(CConnection* connection)
 {
     m_InUse.push_back(connection);
@@ -401,8 +441,10 @@ CDriverContext::MakePooledConnection(const CDBConnParams& params)
 
     // Precondition check.
     if (params.GetServerName().empty() ||
-            params.GetUserName().empty() ||
-            params.GetPassword().empty()) {
+        (!TDbapi_CanUseKerberos::GetDefault()
+         &&  (params.GetUserName().empty()
+              ||  params.GetPassword().empty())))
+    {
         string err_msg("Insufficient info/credentials to connect.");
 
         if (params.GetServerName().empty()) {
@@ -685,6 +727,10 @@ CDriverContext::ReadDBConfParams(const string&  service_name,
         params->flags += SDBConfParams::fIOTimeoutSet;
         params->io_timeout = reg.Get(section_name, "io_timeout");
     }
+    if (reg.HasEntry(section_name, "cancel_timeout", IRegistry::fCountCleared)) {
+        params->flags += SDBConfParams::fCancelTimeoutSet;
+        params->cancel_timeout = reg.Get(section_name, "cancel_timeout");
+    }
     if (reg.HasEntry(section_name, "exclusive_server", IRegistry::fCountCleared)) {
         params->flags += SDBConfParams::fSingleServerSet;
         params->single_server = reg.Get(section_name, "exclusive_server");
@@ -811,6 +857,23 @@ CDriverContext::MakeConnection(const CDBConnParams& params)
                 SetTimeout(NStr::StringToInt(value));
             }
         }
+        if (conf_params.IsCancelTimeoutSet()) {
+            if (conf_params.cancel_timeout.empty()) {
+                SetCancelTimeout(0);
+            }
+            else {
+                SetCancelTimeout(NStr::StringToInt(conf_params.cancel_timeout));
+            }
+        }
+        else {
+            string value(params.GetParam("cancel_timeout"));
+            if (value == "default") {
+                SetCancelTimeout(10);
+            }
+            else if (!value.empty()) {
+                SetCancelTimeout(NStr::StringToInt(value));
+            }
+        }
         if (conf_params.IsSingleServerSet()) {
             if (conf_params.single_server.empty()) {
                 act_params.SetParam("single_server", "true");
@@ -850,10 +913,8 @@ CDriverContext::MakeConnection(const CDBConnParams& params)
         act_params.SetDatabaseName(db_name);
         act_params.SetPassword(password);
 
-        t_con = CDbapiConnMgr::Instance().GetConnectionFactory()->MakeDBConnection(
-                    *this,
-                    act_params);
-
+        CRef<IDBConnectionFactory> factory = CDbapiConnMgr::Instance().GetConnectionFactory();
+        t_con = factory->MakeDBConnection(*this, act_params);
 
         if((!t_con && act_params.GetParam("do_not_connect") == "true")) {
             return NULL;
@@ -861,10 +922,17 @@ CDriverContext::MakeConnection(const CDBConnParams& params)
 
         if (!t_con) {
             string err;
-
             err += "Cannot connect to the server '" + act_params.GetServerName();
             err += "' as user '" + act_params.GetUserName() + "'";
-            DATABASE_DRIVER_ERROR( err, 100011 );
+
+            CDB_ClientEx ex(DIAG_COMPILE_INFO, NULL, err, eDiag_Error, 100011);
+            CDB_UserHandler::TExceptions* expts = factory->GetExceptions();
+            if (expts) {
+                NON_CONST_REVERSE_ITERATE(CDB_UserHandler::TExceptions, it, *expts) {
+                    ex.AddPrevious(*it);
+                }
+            }
+            throw ex;
         }
 
         // Set database ...
@@ -916,17 +984,17 @@ void CDriverContext::SetClientCharset(const string& charset)
     m_ClientCharset = charset;
     m_ClientEncoding = eEncoding_Unknown;
 
-    if (NStr::CompareNocase(charset.c_str(), "UTF-8") == 0 ||
-        NStr::CompareNocase(charset.c_str(), "UTF8") == 0) {
+    if (NStr::CompareNocase(charset, "UTF-8") == 0 ||
+        NStr::CompareNocase(charset, "UTF8") == 0) {
         m_ClientEncoding = eEncoding_UTF8;
-    } else if (NStr::CompareNocase(charset.c_str(), "Ascii") == 0) {
+    } else if (NStr::CompareNocase(charset, "Ascii") == 0) {
         m_ClientEncoding = eEncoding_Ascii;
-    } else if (NStr::CompareNocase(charset.c_str(), "ISO8859_1") == 0 ||
-               NStr::CompareNocase(charset.c_str(), "ISO8859-1") == 0
+    } else if (NStr::CompareNocase(charset, "ISO8859_1") == 0 ||
+               NStr::CompareNocase(charset, "ISO8859-1") == 0
                ) {
         m_ClientEncoding = eEncoding_ISO8859_1;
-    } else if (NStr::CompareNocase(charset.c_str(), "Windows_1252") == 0 ||
-               NStr::CompareNocase(charset.c_str(), "Windows-1252") == 0) {
+    } else if (NStr::CompareNocase(charset, "Windows_1252") == 0 ||
+               NStr::CompareNocase(charset, "Windows-1252") == 0) {
         m_ClientEncoding = eEncoding_Windows_1252;
     }
 }
