@@ -1,4 +1,4 @@
-/*  $Id: reader.cpp 214690 2010-12-07 20:32:09Z vasilche $
+/*  $Id: reader.cpp 406821 2013-07-16 15:12:01Z vasilche $
  * ===========================================================================
  *                            PUBLIC DOMAIN NOTICE
  *               National Center for Biotechnology Information
@@ -123,10 +123,10 @@ void CReader::InitParams(CConfig& conf,
                     DEFAULT_WAIT_TIME_ERRORS);
     m_WaitTime.Init(conf, driver_name, s_WaitTimeParams);
 
-    TConn max_connections = conf.GetInt(driver_name,
-                                        NCBI_GBLOADER_READER_PARAM_NUM_CONN,
-                                        CConfig::eErr_NoThrow,
-                                        -1);
+    int max_connections = conf.GetInt(driver_name,
+                                      NCBI_GBLOADER_READER_PARAM_NUM_CONN,
+                                      CConfig::eErr_NoThrow,
+                                      -1);
     if ( max_connections == -1 ) {
         max_connections = conf.GetInt(driver_name,
                                       NCBI_GBLOADER_READER_PARAM2_NUM_CONN,
@@ -148,7 +148,7 @@ void CReader::OpenInitialConnection(bool force)
                 return;
             }
             catch ( CLoaderException& exc ) {
-                x_ReleaseConnection(conn, true);
+                x_ReleaseClosedConnection(conn);
                 if ( exc.GetErrCode() == exc.eNoConnection ) {
                     // no connection can be opened
                     throw;
@@ -162,7 +162,7 @@ void CReader::OpenInitialConnection(bool force)
                 }
             }
             catch ( CException& exc ) {
-                x_ReleaseConnection(conn, true);
+                x_ReleaseClosedConnection(conn);
                 LOG_POST_X(2, Warning<<"CReader: "
                            "cannot open initial connection: "<<exc.what());
                 if ( attempt >= GetRetryCount() ) {
@@ -314,7 +314,7 @@ void CReader::x_AddConnection(void)
     CMutexGuard guard(m_ConnectionsMutex);
     TConn conn = m_NextNewConnection++;
     x_AddConnectionSlot(conn);
-    x_ReleaseConnection(conn, true);
+    x_ReleaseClosedConnection(conn);
     ++m_MaxConnections;
     _ASSERT(m_MaxConnections > 0);
 }
@@ -337,7 +337,7 @@ CReader::TConn CReader::x_AllocConnection(bool oldest)
     }
     m_NumFreeConnections.Wait();
     CMutexGuard guard(m_ConnectionsMutex);
-    TConnSlot slot;
+    SConnSlot slot;
     if ( oldest ) {
         slot =  m_FreeConnections.back();
         m_FreeConnections.pop_back();
@@ -346,26 +346,44 @@ CReader::TConn CReader::x_AllocConnection(bool oldest)
         slot =  m_FreeConnections.front();
         m_FreeConnections.pop_front();
     }
-    if ( !slot.second.IsEmpty() ) {
-        double age = CTime(CTime::eCurrent).DiffNanoSecond(slot.second)*1e-9;
+    if ( !slot.m_LastUseTime.IsEmpty() ) {
+        double age = CTime(CTime::eCurrent).DiffNanoSecond(slot.m_LastUseTime)*1e-9;
         if ( age > 60 ) {
             // too old, disconnect
-            x_DisconnectAtSlot(slot.first, false);
+            x_DisconnectAtSlot(slot.m_Conn, false);
+        }
+        else if ( age < slot.m_RetryDelay ) {
+            double wait_seconds = slot.m_RetryDelay - age;
+            LOG_POST_X(6, Warning << "CReader: waiting "<<
+                       wait_seconds<<"s before next command");
+            _TRACE("CReader: waiting "<<wait_seconds<<
+                   "s before next connection");
+            SleepMicroSec((unsigned long)(wait_seconds*1e6));
         }
     }
-    return slot.first;
+    return slot.m_Conn;
 }
 
 
-void CReader::x_ReleaseConnection(TConn conn, bool oldest)
+void CReader::x_ReleaseClosedConnection(TConn conn)
 {
     CMutexGuard guard(m_ConnectionsMutex);
-    if ( oldest ) {
-        m_FreeConnections.push_back(TConnSlot(conn, CTime()));
-    }
-    else {
-        m_FreeConnections.push_front(TConnSlot(conn, CTime(CTime::eCurrent)));
-    }
+    SConnSlot slot;
+    slot.m_Conn = conn;
+    slot.m_RetryDelay = 0;
+    m_FreeConnections.push_back(slot);
+    m_NumFreeConnections.Post(1);
+}
+
+
+void CReader::x_ReleaseConnection(TConn conn, double retry_delay)
+{
+    CMutexGuard guard(m_ConnectionsMutex);
+    SConnSlot slot;
+    slot.m_Conn = conn;
+    slot.m_LastUseTime = CTime(CTime::eCurrent);
+    slot.m_RetryDelay = retry_delay;
+    m_FreeConnections.push_front(slot);
     m_NumFreeConnections.Post(1);
 }
 
@@ -375,7 +393,7 @@ void CReader::x_AbortConnection(TConn conn, bool failed)
     CMutexGuard guard(m_ConnectionsMutex);
     try {
         x_DisconnectAtSlot(conn, failed);
-        x_ReleaseConnection(conn, true);
+        x_ReleaseClosedConnection(conn);
         return;
     }
     catch ( exception& exc ) {
@@ -386,7 +404,7 @@ void CReader::x_AbortConnection(TConn conn, bool failed)
     try {
         x_RemoveConnectionSlot(conn);
     }
-    catch ( ... ) {
+    catch ( exception& ) {
         // ignore
     }
     _ASSERT(m_MaxConnections > 0);
@@ -510,8 +528,8 @@ bool CReader::LoadSeq_idTaxId(CReaderRequestResult& result,
 bool CReader::LoadAccVers(CReaderRequestResult& result,
                           const TIds& ids, TLoaded& loaded, TIds& ret)
 {
-    int count = ids.size();
-    for ( int i = 0; i < count; ++i ) {
+    size_t count = ids.size();
+    for ( size_t i = 0; i < count; ++i ) {
         if ( loaded[i] ) {
             continue;
         }
@@ -531,8 +549,8 @@ bool CReader::LoadAccVers(CReaderRequestResult& result,
 bool CReader::LoadGis(CReaderRequestResult& result,
                       const TIds& ids, TLoaded& loaded, TGis& ret)
 {
-    int count = ids.size();
-    for ( int i = 0; i < count; ++i ) {
+    size_t count = ids.size();
+    for ( size_t i = 0; i < count; ++i ) {
         if ( loaded[i] ) {
             continue;
         }
@@ -552,8 +570,8 @@ bool CReader::LoadGis(CReaderRequestResult& result,
 bool CReader::LoadLabels(CReaderRequestResult& result,
                          const TIds& ids, TLoaded& loaded, TLabels& ret)
 {
-    int count = ids.size();
-    for ( int i = 0; i < count; ++i ) {
+    size_t count = ids.size();
+    for ( size_t i = 0; i < count; ++i ) {
         if ( loaded[i] ) {
             continue;
         }
@@ -573,8 +591,8 @@ bool CReader::LoadLabels(CReaderRequestResult& result,
 bool CReader::LoadTaxIds(CReaderRequestResult& result,
                          const TIds& ids, TLoaded& loaded, TTaxIds& ret)
 {
-    int count = ids.size();
-    for ( int i = 0; i < count; ++i ) {
+    size_t count = ids.size();
+    for ( size_t i = 0; i < count; ++i ) {
         if ( loaded[i] ) {
             continue;
         }
@@ -778,7 +796,7 @@ void CReader::SetAndSaveStringSeq_ids(CReaderRequestResult& result,
 
 void CReader::SetAndSaveStringGi(CReaderRequestResult& result,
                                  const string& seq_id,
-                                 int gi) const
+                                 TGi gi) const
 {
     CLoadLockSeq_ids seq_ids(result, seq_id);
     SetAndSaveStringGi(result, seq_id, seq_ids, gi);
@@ -795,7 +813,7 @@ void CReader::SetAndSaveSeq_idSeq_ids(CReaderRequestResult& result,
 
 void CReader::SetAndSaveSeq_idGi(CReaderRequestResult& result,
                                  const CSeq_id_Handle& seq_id,
-                                 int gi) const
+                                 TGi gi) const
 {
     CLoadLockSeq_ids seq_ids(result, seq_id);
     SetAndSaveSeq_idGi(result, seq_id, seq_ids, gi);
@@ -869,7 +887,7 @@ void CReader::SetAndSaveStringSeq_ids(CReaderRequestResult& result,
 void CReader::SetAndSaveStringGi(CReaderRequestResult& result,
                                  const string& seq_id,
                                  CLoadLockSeq_ids& seq_ids,
-                                 int gi) const
+                                 TGi gi) const
 {
     if ( seq_ids->IsLoadedGi() ) {
         return;
@@ -907,7 +925,7 @@ void CReader::SetAndSaveSeq_idSeq_ids(CReaderRequestResult& result,
 void CReader::SetAndSaveSeq_idGi(CReaderRequestResult& result,
                                  const CSeq_id_Handle& seq_id,
                                  CLoadLockSeq_ids& seq_ids,
-                                 int gi) const
+                                 TGi gi) const
 {
     if ( seq_ids->IsLoadedGi() ) {
         return;
@@ -1048,6 +1066,7 @@ CReaderAllocatedConnection::CReaderAllocatedConnection(CReaderRequestResult& res
         m_Conn = reader->x_AllocConnection();
         m_Reader = reader;
         m_Result = &result;
+        result.ClearRetryDelay();
         result.m_AllocatedConnection = this;
     }
     else if ( pconn->m_Reader == reader ) {
@@ -1090,9 +1109,10 @@ void CReaderAllocatedConnection::Release(void)
     if ( m_Result ) {
         _ASSERT(m_Result->m_AllocatedConnection == this);
         _ASSERT(m_Reader);
+        double retry_delay = m_Result->GetRetryDelay();
         m_Result->m_AllocatedConnection = 0;
         m_Result = 0;
-        m_Reader->x_ReleaseConnection(m_Conn);
+        m_Reader->x_ReleaseConnection(m_Conn, min(retry_delay, 60.0));
     }
 }
 
